// Copyright (c) 2016 nyorain 
// Distributed under the Boost Software License, Version 1.0.
// See accompanying file LICENSE or copy at http://www.boost.org/LICENSE_1_0.txt

//TODO: some of those functions are not correctly implemented...

#pragma once

#ifndef NYTL_INCLUDE_TRANSFORM_INL
#define NYTL_INCLUDE_TRANSFORM_INL

#ifdef DOXYGEN
namespace nytl{
#endif

namespace detail
{

//returns the plane (e.g. 0,1 for xy or 2,3 for zw) for a given id in a given dimension
inline Vec2<std::size_t> indexPlane(std::size_t dim, std::size_t idx)
{
	auto ret = Vec2<std::size_t> {};

	while(idx >= dim - 1)
	{
		ret[0]++;
		idx -= (dim - 1);
		dim--;
	}

	ret[1] = idx + 1;

	return ret;
}

}

//basic ops
///\relates Mat Transform
template<std::size_t D, typename P>
void scale(SquareMat<D, P>& mat, const Vec<D - 1, P>& scal)
{
	for(std::size_t i(0); i < D - 1; ++i)
		mat[i][i] *= scal[i];
}

///\relates Mat Transform
template<std::size_t D, typename P>
void translate(SquareMat<D, P>& mat, const Vec<D - 1, P>& trans)
{
	for(std::size_t i(0); i < D - 1; ++i)
		mat[i][D - 1] += trans[i];
}

///\relates Mat Transform
template<typename P>
void rotate(SquareMat<3, P>& mat, const Vec<1, P>& planeRot)
{
	auto rotMat = identityMat<3>();

	auto c = std::cos(planeRot[0]);
	auto s = std::sin(planeRot[0]);

	rotMat[0][0] = c;
	rotMat[0][1] = -s;
	rotMat[1][0] = s;
	rotMat[1][1] = c;

	mat *= rotMat;
}

template<typename P>
void rotate(SquareMat<4, P>& mat, const Vec3<P>& vec, const P& angle)
{
	const P c = std::cos(angle);
	const P s = std::sin(angle);

	Vec3<P> axis = normalize(vec);
	Vec3<P> temp = (P(1) - c) * axis;

	auto rotate = identityMat<4, P>();
	rotate[0][0] = c + temp[0] * axis[0];
	rotate[0][1] = 0 + temp[0] * axis[1] + s * axis[2];
	rotate[0][2] = 0 + temp[0] * axis[2] - s * axis[1];

	rotate[1][0] = 0 + temp[1] * axis[0] - s * axis[2];
	rotate[1][1] = c + temp[1] * axis[1];
	rotate[1][2] = 0 + temp[1] * axis[2] + s * axis[0];

	rotate[2][0] = 0 + temp[2] * axis[0] + s * axis[1];
	rotate[2][1] = 0 + temp[2] * axis[1] - s * axis[0];
	rotate[2][2] = c + temp[2] * axis[2];
/*
	mat[0] = mat[0] * rotate[0][0] + mat[1] * rotate[0][1] + mat[2] * rotate[0][2];
	mat[1] = mat[0] * rotate[1][0] + mat[1] * rotate[1][1] + mat[2] * rotate[1][2];
	mat[2] = mat[0] * rotate[2][0] + mat[1] * rotate[2][1] + mat[2] * rotate[2][2];
*/
	mat *= rotate;
}

template<typename P>
void rotate(SquareMat<4, P>& mat, const Vec3<P>& vec)
{
	rotate(mat, Vec3<P>(normalize(vec)), P(length(vec)));
}

//copy helper
///\relates Mat Transform
template<std::size_t D, typename P>
SquareMat<D, P> scaleCopy(SquareMat<D, P> mat, const Vec<D, P>& scal)
{
	scale(mat, scal);
	return mat;
}

///\relates Mat Transform
template<std::size_t D, typename P>
SquareMat<D, P> translateCopy(SquareMat<D, P> mat, const Vec<D, P>& trans)
{
	translate(mat, trans);
	return mat;
}

///\relates Mat Transform
template<std::size_t D, typename P>
SquareMat<D, P> rotateCopy(SquareMat<D, P> mat, const Vec<D, P>& axes, const P& angle)
{
	rotate(mat, axes, angle);
	return mat;
}

///\relates Mat Transform
template<std::size_t D, typename P>
SquareMat<D, P> rotateCopy(SquareMat<D, P> mat, const Vec<D, P>& rot)
{
	rotate(mat, rot);
	return mat;
}

template<typename P>
SquareMat<4, P> frustrum3(P left, P right, P top, P bottom, P pnear, P far)
{
	auto ret = SquareMat<4, P>(0);

	ret[0][0] = (P(2) * pnear) / (right - left);
	ret[1][1] = (P(2) * pnear) / (top - bottom);
	ret[2][2] = -(far * pnear) / (far - pnear);

	ret[0][2] = (right + left) / (right - left);
	ret[1][2] = (top + bottom) / (top - bottom);
	ret[3][2] = -1;

	ret[2][3] = (P(-2) * far * pnear) / (far - pnear);
	return ret;
}

template<typename P>
SquareMat<4, P> frustum3Symmetrical(P width, P height, P pnear, P pfar)
{
	return perspective3(-width / P(2), width / P(2), height / P(2), -height / P(2), pnear, pfar);
}

template<typename P>
SquareMat<4, P> perspective3(P fov, P aspect, P pnear, P pfar)
{
	P const f = P(1) / std::tan(fov / P(2));

	auto ret = Mat4<P>(0);
	ret[0][0] = f / aspect;
	ret[1][1] = f;

	ret[2][2] = -(pfar + pnear) / (pfar - pnear);
	ret[3][2] = -1;

	ret[2][3] = -(P(2) * pfar * pnear) / (pfar - pnear);
	return ret;
}


template<typename P>
SquareMat<4, P> ortho3(P left, P right, P top, P bottom, P pnear, P far)
{
	auto ret = Mat4<P>(0);

	ret[0][0] = P(2) / (right - left);
	ret[1][1] = P(2) / (top - bottom);
	ret[2][2] = P(2) / (far - pnear);

	ret[0][3] = - ((right + left) / (right - left));
	ret[1][3] = - ((top + bottom) / (top - bottom));
	ret[2][3] = - ((far + pnear) / (far - pnear));
	ret[3][3] = 1;
	return ret;
}

template<typename P>
SquareMat<4, P> ortho3(P width, P height, P pnear, P pfar)
{
	return ortho3(-width / P(2), width / P(2), height / P(2), -height / P(2), pnear, pfar);
}

template<typename P>
SquareMat<4, P> lookAt(const Vec3<P>& eye, const Vec3<P>& center, const Vec3<P>& up)
{
	const auto z = normalize(center - eye); //z
	const auto x = normalize(cross(z, up)); //x
	const auto y = cross(x, z); //y

	auto ret = identityMat<4, P>();

	ret.row(0) = x;
	ret.row(1) = y;
	ret.row(2) = -z;

	ret[0][3] = -dot(x, eye);
	ret[1][3] = -dot(y, eye);
	ret[2][3] = dot(z, eye);

	return ret;
}

#ifdef DOXYGEN
}
#endif

#endif //header guard
